==> kite-1.0b1.tar.gz: kite-1.0b1/modules/System/vm/thread.c <==
/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include "kite_object.h"
#include "vm/kite_vm.h"
        
KITE_CLASS_METHOD(Thread_getCurrent) {
    kite_object_t *class, *ret;
    
    class = kite_dereference_and_load(thd, "System.vm.thread");
    ret = kite_new_instance(thd, class);
    ret->builtin_data.threadvalue = thd;
    
    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(Thread_new) {
    kite_object_t *class, *ret;
    kite_thread_t *newthd;
    kite_object_t *method, *methodargs, *exc;
    
    method = args->builtin_data.listvalue.car;
    methodargs = args->builtin_data.listvalue.cdr;
    methodargs = methodargs->builtin_data.listvalue.car;
    
    if (method->type != OBJ_METHOD ||
        methodargs->type != OBJ_LIST) {
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch",
                                 "Operand types do not match (method, list).");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
    }
    
    newthd = kite_new_thread_without_start(thd->vm);
    kite_push_stack(&newthd->running_stack, FALSE, kite_reference_object(methodargs));
    kite_start_bytecode(newthd, method);
    
    class = kite_dereference_and_load(thd, "System.vm.thread");
    ret = kite_new_instance(thd, class);
    ret->builtin_data.threadvalue = newthd;
    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(Thread_join) {
    kite_object_t *arg = this;
    kite_thread_t *joinee = arg->builtin_data.threadvalue;
    
    if (joinee != NULL) kite_join_thread(thd->vm, joinee);
}

KITE_CLASS_METHOD(Thread_exit) {
    kite_object_t *arg = args->builtin_data.listvalue.car;
    
    if (arg == NULL) {
        pthread_exit(NULL);
    }
    
    kite_thread_t *joinee = arg->builtin_data.threadvalue;
    kite_exit_thread(thd->vm, joinee);
}

KITE_MODULE_INITIALIZER(thread) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "thread");

    /* Add methods. */
    kite_add_method(thread, newclass, "getCurrent", 
        kite_new_method(thread, FUNC_COMPILED, Thread_getCurrent, 0));
    kite_add_method(thread, newclass, "new", 
        kite_new_method(thread, FUNC_COMPILED, Thread_new, 2));
    kite_add_method(thread, newclass, "join", 
        kite_new_method(thread, FUNC_COMPILED, Thread_join, 0));
    kite_add_method(thread, newclass, "exit", 
        kite_new_method(thread, FUNC_COMPILED, Thread_exit, 0));
    kite_add_method(thread, newclass, "exit", 
        kite_new_method(thread, FUNC_COMPILED, Thread_exit, 1));
        
    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "thread", 1);
    kite_set_property(thread, parent, "thread", newclass);
}

==> kite_1.0.2.orig.tar.gz: kite-1.0.2/modules/System/vm/thread.c <==
/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include "kite_object.h"
#include "vm/kite_vm.h"
        
KITE_CLASS_METHOD(Thread_getCurrent) {
    kite_object_t *class, *ret;
    
    KITE_NO_ARGS;
    (void)this;
    
    class = kite_dereference_and_load(thd, "System.vm.thread");
    ret = kite_new_instance(thd, class);
    ret->builtin_data.threadvalue = thd;
    
    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(Thread_new) {
    kite_object_t *class, *ret;
    kite_thread_t *newthd;
    kite_object_t *method, *methodargs, *exc;
    
    (void)this;
    
    method = args->builtin_data.listvalue.car;
    methodargs = args->builtin_data.listvalue.cdr;
    methodargs = methodargs->builtin_data.listvalue.car;
    
    if (method->type != OBJ_METHOD ||
        methodargs->type != OBJ_LIST) {
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch",
                                 "Operand types do not match (method, list).");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }
    
    newthd = kite_new_thread_without_start(thd->vm);
    kite_push_stack(&newthd->running_stack, FALSE, kite_reference_object(methodargs));
    kite_start_bytecode(newthd, method);
    
    class = kite_dereference_and_load(thd, "System.vm.thread");
    ret = kite_new_instance(thd, class);
    ret->builtin_data.threadvalue = newthd;
    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(Thread_join) {
    kite_object_t *arg = this;
    kite_thread_t *joinee = arg->builtin_data.threadvalue;
    
    KITE_NO_ARGS;
    if (joinee != NULL) kite_join_thread(thd->vm, joinee);
}

KITE_CLASS_METHOD(Thread_exit) {
    kite_object_t *arg = args->builtin_data.listvalue.car;
    kite_thread_t *joinee;

    (void)this;
    KITE_NO_ARGS;
    
    if (arg == NULL) {
#ifndef WIN32
        pthread_exit(NULL);
#else
        ExitThread(0);
#endif /* WIN32 */
    }
    
    joinee = arg->builtin_data.threadvalue;
    kite_exit_thread(thd->vm, joinee);
}

KITE_MODULE_INITIALIZER(thread) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "thread");
    kite_set_docstring(newclass, "User-level access to the Kite thread system.");
    
    /* Add methods. */
    kite_add_method(thread, newclass, "getCurrent", 
        kite_new_method_compiled_with_docs(thread, Thread_getCurrent, 
                                  "Returns the current thread.", 0));
    kite_add_method(thread, newclass, "new", 
        kite_new_method_compiled_with_docs(thread, Thread_new, 
                                  "Creates a new thread.", 2,
                                  "method", "The method to run in the new thread.",
                                  "args", "The arguments to pass to the method."));
    kite_add_method(thread, newclass, "join", 
        kite_new_method_compiled_with_docs(thread, Thread_join, 
                                  "Wait for given thread to terminate.", 0));
    kite_add_method(thread, newclass, "exit", 
        kite_new_method_compiled_with_docs(thread, Thread_exit, 
                                  "Exits current thread.", 0));
    kite_add_method(thread, newclass, "exit", 
        kite_new_method_compiled_with_docs(thread, Thread_exit, 
                                  "Exits the given thread", 1,
                                  "thd", "The thread to exit."));
        
    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "thread", 1, NULL);
    kite_set_property(thread, parent, "thread", newclass);
}
